# Generated by Django 5.0.4 on 2024-07-09 14:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0023_remove_container_cpus_remove_project_max_containers_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='container',
            name='username',
            field=models.CharField(default='root', help_text='Username for the RStudio Rocker container.', max_length=255),
        ),
    ]



import subprocess

from django.conf import settings
from podman import PodmanClient
from podman.errors import NotFound

from .models import Container, Project

# Get the podman settings from the Django settings
PODMAN_HOST_ADDRESS = settings.PODMAN_HOST_ADDRESS
PODMAN_SOCKET = settings.PODMAN_SOCKET

# These are versions of the rocker/rstudio image that recognize that they
# are in a rootless podman container and adjust the paths accordingly. The
# feature was applied in this commit:
# https://github.com/rocker-org/rocker-versioned2/commit/8ab4e7d01acaf3070f4628251897b779b09ca1a2
SUPPORTED_RSTUDIO_VERSIONS_WITH_ROOT_PATCH = [
    # 4.1
    "4.4.1",
    "4.4.0",
    # 4.3
    "4.3.2",
    "4.3.1",
    "4.3.0",
    # 4.2
    "4.2.3",
    "4.2.2",
    "4.2.1",
    "4.2.0",
    # 4.1
    "4.1.3",
    # 4.0
    "4.0.5",
]

# These are the version where the user is still 'rstudio' and not 'root'
SUPPORTED_RSTUDIO_VERSIONS_WITHOUT_ROOT_PATCH = [
    # 4.1
    "4.1.2",
    "4.1.1",
    "4.1.0",
    # 4.0
    "4.0.4",
    "4.0.3",
    "4.0.2",
    "4.0.1",
    "4.0.0",
]

# 'container' always refers to the Container model instance whereas
# 'podman_container' refers to the podman container instance


class PodmanError(Exception):
    """Base class for Podman exceptions."""

    pass


def get_podman_container(client, project, container):
    """Get the podman container object if it exists, else return None."""
    try:
        return client.containers.get(f"rstudio_{project.slug}_{container.container_id}")
    except NotFound:
        return None


def start_and_save_container(podman_container, container):
    """Start the podman container and save the container details."""
    podman_container.start()
    podman_container.reload()

    container.port = podman_container.ports["8787/tcp"][0]["HostPort"]
    container.local_url = f"http://{PODMAN_HOST_ADDRESS}:{container.port}"
    container.save()

    print(f"Container started at {container.local_url}")


def create_rstudio_container_config(project, container, password):
    """Create the container configuration. Right now, it only supports RStudio."""

    # Specify the image, name, and other configurations for the container
    config = {
        "image": f"{container.image_host}/{container.image}:{container.tag}",
        "name": f"rstudio_{project.slug}_{container.container_id}",
        "detach": True,
        "ports": {"8787/tcp": container.port},  # Assuming default RStudio port
        "environment": {
            "PASSWORD": password,  # Set the PASSWORD environment variable for RStudio
            "ROOT": "TRUE",
        },
        "mem_limit": f"{container.ram}g",
        # Restart the container if it stops
        "restart_policy": {"Name": "always"},
        "mounts": [
            {
                "type": "bind",
                "source": f"{project.data_path}",
                "target": "/data",
                "read_only": False,
            },
        ],
    }

    # If user requests a version with the root patch, mount the workspace to
    # the /root directory
    if container.tag in SUPPORTED_RSTUDIO_VERSIONS_WITH_ROOT_PATCH:
        config["mounts"].append(
            {
                "type": "bind",
                "source": f"{project.workspace_path}",
                "target": f"/root/{project.slug}",
                "read_only": False,
            },
        )
        # Set the username to root, so it can be displayed in the UI
        container.username = "root"
        container.save()
        return config

    # If user requests a version without the root patch, mount the workspace
    # to the /home/rstudio directory
    elif container.tag in SUPPORTED_RSTUDIO_VERSIONS_WITHOUT_ROOT_PATCH:
        (
            config["mounts"].append(
                {
                    "type": "bind",
                    "source": f"{project.workspace_path}",
                    "target": f"/home/rstudio/{project.slug}",
                    "read_only": False,
                },
            ),
        )
        # Set the username to rstudio, so it can be displayed in the UI
        container.username = "rstudio"
        container.save()
        return config

    # Force user to specify a supported version
    else:
        raise PodmanError(
            "Unsupported RStudio version. Only versions 4.x.x are currently supported."
        )


def start_container(project: Project, container: Container, password: str = None):
    """Start the container."""
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        podman_container = get_podman_container(client, project, container)

        if podman_container is not None:
            print("Container already exists. No need to create.")
            if podman_container.status != "running":
                print("Starting container...")
                start_and_save_container(podman_container, container)
        else:
            print("Container does not exist creating...")
            container_config = create_rstudio_container_config(
                project, container, password
            )
            try:
                client.images.pull(container_config["image"])
                podman_container = client.containers.create(**container_config)
                start_and_save_container(podman_container, container)
            except Exception as e:
                # TODO Handle cases where the container creation fails because
                # of a missing directory and create it during project creation
                # ideally
                raise PodmanError(f"Failed to start container: {e}")


def stop_container(project: Project, container: Container):
    """Stop the container."""
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            podman_container = client.containers.get(
                f"rstudio_{project.slug}_{container.container_id}"
            )
            podman_container.stop()
        except Exception as e:
            raise PodmanError(f"Failed to stop container: {e}")


def remove_container(project: Project, container: Container):
    """Remove the container."""
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            podman_container = client.containers.get(
                f"rstudio_{project.slug}_{container.container_id}"
            )
            podman_container.remove(force=True)
            print("Container removed.")
        except Exception as e:
            raise PodmanError(f"Failed to remove container: {e}")


def prune_containers():
    """Prune all stopped containers."""
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            client.containers.prune()
            print("Containers pruned.")
        except Exception as e:
            raise PodmanError(f"Failed to prune containers: {e}")


def is_container_running(project: Project, container: Container):
    """Check if the container is running."""
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            podman_container = client.containers.get(
                f"rstudio_{project.slug}_{container.container_id}"
            )
            return podman_container.status == "running"
        except Exception as e:
            raise PodmanError(f"Failed to get container status: {e}")


def update_container_ram(project: Project, container: Container):
    """Update the RAM allocation for the container."""
    try:
        cmd = [
            "podman",
            "container",
            "update",
            f"rstudio_{project.slug}_{container.container_id}",
            "--memory",
            f"{container.ram}g",
        ]
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as e:
        raise PodmanError(f"Failed to update container RAM allocation: {e}")



# Generated by Django 5.0.4 on 2024-07-02 14:19

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0021_remove_container_tailscale_funnel_url_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='container',
            name='image_host',
            field=models.CharField(default='docker.io', help_text='The name of the host where the image should be pulled from', max_length=255),
        ),
        migrations.AlterField(
            model_name='project',
            name='git_repo_url',
            field=models.CharField(blank=True, help_text='URL of the Git repository. You can use SSH or HTTPS URLs. For example, you could use forgejo@code.osi.uni-mannheim.de:r-shepard/<repo_name>.git', max_length=255, null=True, validators=[django.core.validators.RegexValidator('^(forgejo@|git@|http:\\/\\/|https:\\/\\/).*(\\.git)$', 'Enter a valid Git URL.')], verbose_name='Git Repository URL'),
        ),
    ]



import os

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django_otp.decorators import otp_required

from r_shepard.api.models import Container, Project
from r_shepard.api.podman import (
    PodmanError,
    is_container_running,
    remove_container,
    start_container,
    stop_container,
    update_container_ram,
)


@login_required
@otp_required
def create_container_view(request, project_pk):
    """Create a new container."""
    if request.method == "POST":
        # Initialize message HTML
        message_html = render_to_string("messages.html")

        # Get project from URL
        project = get_object_or_404(Project, pk=project_pk)

        # Get parameters from POST request
        password = request.POST.get("password")
        ram = int(request.POST.get("ram"))
        tag = str(request.POST.get("tag"))
        tag = tag or "4.4.1"

        # Get system RAM
        system_ram = (
            os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES") / (1024.0**3)
        )

        # Get RAM used by currently running containers
        used_ram = sum(
            container.ram
            for container in project.containers.all()
            if container.is_running
        )

        # Check if there are enough resources available for the container
        if ram > system_ram - used_ram - 1:
            messages.error(request, "Not enough RAM available.")
            message_html = render_to_string(
                "messages.html", {"messages": messages.get_messages(request)}
            )
        else:
            container = Container.objects.create(
                project=project, tag=tag, password=password, ram=ram
            )
            try:
                start_container(project, container, password)
                container.is_running = True
                container.save()
            except PodmanError as e:
                container.delete()
                messages.error(request, str(e))
                message_html = render_to_string(
                    "messages.html", {"messages": messages.get_messages(request)}
                )

        container_list_html = render_to_string(
            "container_list.html", {"project": project}
        )

        return HttpResponse(container_list_html + message_html)


@login_required
@otp_required
def start_container_view(request, project_pk, container_pk):
    """Start a container."""
    if request.method == "POST":
        # Initialize message HTML
        message_html = render_to_string("messages.html")

        # Get project and container from URL
        project = get_object_or_404(Project, pk=project_pk)
        container = get_object_or_404(Container, pk=container_pk)

        # Start container if it is not running
        try:
            if not is_container_running(project, container):
                start_container(project, container)
        except PodmanError as e:
            messages.error(request, str(e))
            message_html = render_to_string(
                "messages.html", {"messages": messages.get_messages(request)}
            )

        # Update container status in the database
        container.is_running = True
        container.save()

        # Render the container list template with the updated project
        container_list_html = render_to_string(
            "container_list.html", {"project": project}
        )

        return HttpResponse(container_list_html + message_html)


@login_required
@otp_required
def stop_container_view(request, project_pk, container_pk):
    """Stop a container."""
    if request.method == "POST":
        # Initialize message HTML
        message_html = render_to_string("messages.html")

        # Get project and container from URL
        project = get_object_or_404(Project, pk=project_pk)
        container = get_object_or_404(Container, pk=container_pk)

        # Stop container if it is running
        try:
            if is_container_running(project, container):
                stop_container(project, container)
        except PodmanError as e:
            messages.error(request, str(e))
            message_html = render_to_string(
                "messages.html", {"messages": messages.get_messages(request)}
            )

        # Update container status in the database
        container.is_running = False
        container.save()

        # Render the container list template with the updated project
        container_list_html = render_to_string(
            "container_list.html", {"project": project}
        )

        return HttpResponse(container_list_html + message_html)


@login_required
@otp_required
def delete_container_view(request, project_pk, container_pk):
    """Delete a container."""
    if request.method == "POST":
        # Initialize message HTML
        message_html = render_to_string("messages.html")

        # Get project and container from URL
        project = get_object_or_404(Project, pk=project_pk)
        container = get_object_or_404(Container, pk=container_pk)

        # Remove container if it is running
        try:
            if is_container_running(project, container):
                remove_container(project, container)
        except PodmanError as e:
            messages.error(request, str(e))
            message_html = render_to_string(
                "messages.html", {"messages": messages.get_messages(request)}
            )

        # Delete container in the database
        container.delete()

        # Render the container list template with the updated project
        container_list_html = render_to_string(
            "container_list.html", {"project": project}
        )

        return HttpResponse(container_list_html + message_html)


@login_required
@otp_required
def change_ram_limit(request, project_pk, container_pk):
    """Temporarily change a container's RAM limit not exceeding what is
    available on the host."""
    # Initialize message HTML
    message_html = render_to_string("messages.html")

    # Get project and container from URL
    project = get_object_or_404(Project, pk=project_pk)
    container = get_object_or_404(Container, pk=container_pk)

    # Get system resources
    system_ram = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES") / (1024.0**3)

    # Get RAM used by currently running containers
    used_ram = sum(
        container.ram for container in project.containers.all() if container.is_running
    )

    # Get new RAM limit from POST request
    new_ram = int(request.POST.get("new_ram"))

    # Check if there is enough RAM available for the new limit
    if new_ram > system_ram - used_ram - 1:
        messages.error(request, "Not enough RAM available.")
        message_html = render_to_string(
            "messages.html", {"messages": messages.get_messages(request)}
        )
    else:
        container.ram = new_ram
        update_container_ram(project, container)
        container.save()

    # Render the container list template with the updated project
    container_list_html = render_to_string("container_list.html", {"project": project})

    return HttpResponse(container_list_html + message_html)



# Generated by Django 5.0.4 on 2024-07-09 09:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0022_container_image_host_alter_project_git_repo_url'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='container',
            name='cpus',
        ),
        migrations.RemoveField(
            model_name='project',
            name='max_containers',
        ),
        migrations.RemoveField(
            model_name='project',
            name='max_cpus',
        ),
        migrations.RemoveField(
            model_name='project',
            name='max_ram',
        ),
    ]



import os

from django.contrib.auth.decorators import login_required
from django.shortcuts import get_object_or_404, render
from django_otp.decorators import otp_required

from r_shepard.api.models import Project


@login_required
@otp_required
def project_list_view(request):
    """Display a list of all projects."""
    projects = Project.objects.all()

    # Get system RAM
    system_ram = round(
        os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES") / (1024**3), 2
    )

    # Get RAM used by currently running containers
    used_ram = sum(
        container.ram
        for project in projects
        for container in project.containers.all()
        if container.is_running
    )

    # Calculate the ratio of used RAM to system RAM
    context = {
        "projects": projects,
        "system_ram": system_ram - 1,
        "used_ram": used_ram,
        "ratio_used_ram_to_system_ram": used_ram / (system_ram - 1) * 100,
    }

    return render(request, "project_list.html", context)


@login_required
@otp_required
def project_detail_view(request, pk):
    """Display the details of a project."""
    project = get_object_or_404(Project, pk=pk)

    if request.method == "POST":
        # Handle POST request here
        pass

    # Logic to stop all podman containers for the project
    def stop_all_project_containers():
        pass

    context = {
        "project": project,
    }

    return render(request, "project_detail.html", context)



import os
from typing import Any

from django import forms
from django.contrib import admin, messages
from django.db.models.query import QuerySet
from django.http import HttpRequest, HttpResponseRedirect
from django.http.response import HttpResponse
from django.urls import reverse

from r_shepard.api.models import Container, Project
from r_shepard.api.podman import (
    PodmanError,
    remove_container,
    start_container,
)

admin.site.site_header = "R-Shepard"
admin.site.site_title = "R-Shepard"
admin.site.index_title = "Admin Area"


class EditContainerForm(forms.ModelForm):
    class Meta:
        model = Container
        fields = "__all__"
        # These fields cannot be edited and are excluded from the edit form
        exclude = [
            "project",
            "container_id",
            "password",
            "port",
            "local_url",
            "is_running",
        ]


class AddContainerForm(forms.ModelForm):
    password = forms.CharField(widget=forms.PasswordInput)

    class Meta:
        model = Container
        fields = "__all__"
        exclude = [
            "container_id",
            "is_running",
            "port",
            "local_url",
        ]


class ContainerAdmin(admin.ModelAdmin):
    def get_readonly_fields(self, request: Any, obj: Any):
        if obj:  # Edit
            return ["project", "container_id"]
        else:  # Add
            return ["container_id", "image", "tag"]

    def get_form(self, request, obj=None, **kwargs):
        if obj:
            return EditContainerForm
        else:
            return AddContainerForm

    def save_model(self, request: Any, obj: Any, form: Any, change: Any) -> None:
        """Called to save or update an object. Overridden to start the container
        via Podman."""

        # Get system RAM
        system_ram = (
            os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES") / (1024.0**3)
        )

        # Get  RAM used by currently running containers
        used_ram = sum(
            container.ram
            for container in obj.project.containers.all()
            if container.is_running
        )

        # Check if there are enough RAM available for the container
        if obj.ram > system_ram - used_ram - 1:
            messages.error(request, "Not enough RAM available.")
            return

        # Start container
        try:
            start_container(obj.project, obj, obj.password)
            obj.is_running = True
        except PodmanError as e:
            messages.error(request, str(e))
            return
        super().save_model(request, obj, form, change)

    def delete_model(self, request: HttpRequest, obj: Container) -> None:
        """
        Called to delete an object. Overridden to ensure the container is
        removed via Podman before being deleted from the database.
        """
        try:
            # Attempt to remove the container using Podman
            remove_container(obj.project, obj)
        except PodmanError as e:
            # If there's an error, display a message but proceed with deletion
            messages.error(request, f"Failed to remove container via Podman: {e}")

        # Proceed with the original deletion process
        super().delete_model(request, obj)

    def delete_queryset(self, request: HttpRequest, queryset: QuerySet[Any]) -> None:
        """
        Called to bulk delete objects. Overridden to ensure each container is
        removed via Podman before being deleted from the database.
        """
        for obj in queryset:
            try:
                # Attempt to remove each container using Podman
                remove_container(obj.project, obj)
            except PodmanError as e:
                # If there's an error, display a message but proceed with deletion
                messages.error(
                    request, f"Failed to remove container {obj.name} via Podman: {e}"
                )

        # Proceed with the original bulk deletion process
        super().delete_queryset(request, queryset)

    # Redirect to project detail page after adding a container
    def response_add(self, request, obj, post_url_continue=None):
        return HttpResponseRedirect(reverse("project_detail", args=[obj.project.pk]))

    def response_change(self, request: HttpRequest, obj: Any) -> HttpResponse:
        return HttpResponseRedirect(reverse("project_detail", args=[obj.project.pk]))


class ProjectAddForm(forms.ModelForm):
    class Meta:
        model = Project
        fields = [
            "name",
            "description",
            "auto_commit_enabled",
            "git_repo_url",
            "commit_interval",
        ]


class ProjectChangeForm(forms.ModelForm):
    class Meta:
        model = Project
        fields = "__all__"


class ProjectAdmin(admin.ModelAdmin):
    def get_form(self, request, obj=None, **kwargs):
        if obj:
            return ProjectChangeForm
        else:
            return ProjectAddForm

    def response_add(self, request, obj, post_url_continue=None):
        return HttpResponseRedirect(reverse("project_list"))

    def response_change(self, request: HttpRequest, obj: Any) -> HttpResponse:
        return HttpResponseRedirect(reverse("project_detail", args=[obj.pk]))


admin.site.register(Container, ContainerAdmin)
admin.site.register(Project, ProjectAdmin)
